'use strict';

var debug = require('debug')('ifnode:application'),
    path = require('path'),
    express = require('express'),

    helper = require('./helper'),
    log = require('./extensions/log'),

    Application = function(options) {
        if(!(this instanceof Application)) {
            return new Application(options);
        }

        _initialize.call(this, options);
        
        Object.freeze(this);
    },
    
    _initialize = function(app_config) {
        if(app_config.alias && typeof app_config.alias !== 'string') {
            log.error('application', 'Alias must be String');
        }

        this.id = require('./helpers/uid')();
        this.alias = app_config.alias || this.id;

        this.project_folder = this.projectFolder =
            app_config.project_folder ||
            app_config.projectFolder ||
            path.dirname(process.argv[1]);

        this.backend_folder = this.backendFolder =
            path.resolve(this.project_folder, 'protected/');

        _initialize_config.call(this, app_config.env || app_config.environment);
        _initialize_listener.call(this);
        _initialize_server.call(this);
        
        return this;
    },

    _initialize_config = function(environment) {
        var config_path;

        if(environment) {
            config_path = path.resolve(this.project_folder, 'config/', environment);
        }
        
        this.config = require('./ConfigBuilder')({
            environment: environment,
            project_folder: this.project_folder,
            backend_folder: this.backend_folder,
            config_path: config_path
        });
    },
    _initialize_listener = function() {
        var app = express(),
            config = this.config,
            app_config = config.application,

            middleware_configs = app_config.middleware,
            express_configs = app_config.express,

            rest = require('./middleware/rest');

        app.use(rest.response());
        if(middleware_configs) {
            this._initialize_middleware(middleware_configs, app);
            app.use(rest.request());
        }

        Object.keys(express_configs).forEach(function(express_option) {
            app.set(express_option, express_configs[express_option]);
        });

        this.listener = app;
    },
    _initialize_server = function() {
        this.server = new (require('./../plugins/ifnode-http'))(
            require('lodash/defaultsDeep')(
                { listener: this.listener },
                this.config.site.local
            )
        );
    };

Application.fn = Application.prototype;

require('./application/middleware')(Application);
require('./application/extensions')(Application);
require('./application/components')(Application);
require('./application/models')(Application);
require('./application/controllers')(Application);


Application.fn.register = function(module) {
    var type_of = typeof module;

    if(
        type_of === 'string' ||
        Array.isArray(module) ||
        (type_of !== 'undefined' && type_of !== 'number')
    ) {
        this._modules = helper.to_array(module);

        return this;
    }

    log.error('plugins', 'Wrong plugin type');
};

Application.fn.load = function() {
    var app = this,

        initialize_models = function() {
            var type = 'schema',
                SchemaFactory = require('./SchemaFactory'),
                modules = app._modules,

                i,
                Schema,
                module;

            for(i = 0; i < modules.length; ++i) {
                module = modules[i][type];

                if(module) {
                    Schema = SchemaFactory();
                    module(app, Schema);
                    app.attach_schema(Schema);
                }
            }

            app._init_models();
        },
        initialize_components = function() {
            var type = 'component',
                Component = app.Component.bind(app),
                modules = app._modules,

                i, module;

            app._components = {};
            app._initialize_components();

            for(i = 0; i < modules.length; ++i) {
                module = modules[i][type];

                if(module) {
                    module(app, Component);
                }
            }

            app._attach_components();
        },
        initialize_controllers = function() {
            var type = 'controller',
                Controller = require('./controller'),
                modules = app._modules,

                i, module;

            for(i = 0; i < modules.length; ++i) {
                module = modules[i][type];

                if(module) {
                    module(app, Controller);
                }
            }

            app._init_controllers();
        },

        initialize_modules = function() {
            var modules = app._modules;

            modules = helper.to_array(modules).map(function(module) {
                return typeof module === 'string'? require_module(module) : module;
            });

            modules.push(require('./../plugins/ifnode-virtual'));

            app._modules = modules;
        },

        require_module = function(module_name) {
            var module;

            try {
                module = require(module_name);
            } catch(e) {
                module = app.ext(module_name);
            }

            return module;
        };

    initialize_modules();
    initialize_models();
    initialize_components();
    initialize_controllers();

    this._is_loaded = true;

    return this;
};
Application.fn.run = function(callback) {
    if(!this._is_loaded) {
        this.load();
    }
    
    this.server.start(callback);
};

Application.fn.down = function(callback) {
    this.server.stop(callback);
};

// helper.define_properties(Application.fn, {
//     'components':  function() { return this._components },
//     'models':      function() { return this._models },
//     'controllers': function() { return this._controllers }
// });

module.exports = Application;
